from webthing import (MultipleThings, Property, Thing, Value, WebThingServer)
import logging
import tornado.ioloop
from ahoy_dtu_webthing.dtu import Dtu, Inverter



class InverterWebThing(Thing):

    # regarding capabilities refer https://iot.mozilla.org/schemas
    # there is also another schema registry http://iotschema.org/docs/full.html not used by webthing
    def __init__(self, description: str, inverter: Inverter):
        Thing.__init__(
            self,
            'urn:dev:ops:Inverter-1',
            ('Inverter ' + inverter.name).strip(),
            ['MultiLevelSensor'],
            description
        )

        self.inverter = inverter

        self.timestamp_last_success = Value("")
        self.add_property(
            Property(self,
                     'timestamp_last_success',
                     self.timestamp_last_success,
                     metadata={
                         'title': 'timestamp_last_success',
                         "type": "string",
                         'description': 'The last update [ISO DateTime]',
                         'readOnly': True,
                     }))

        self.uri = Value(inverter.uri)
        self.add_property(
            Property(self,
                     'uri',
                     self.uri,
                     metadata={
                         'title': 'uri',
                         "type": "string",
                         'description': 'The inverter uri',
                         'readOnly': True,
                     }))

        self.name = Value(inverter.name)
        self.add_property(
            Property(self,
                     'name',
                     self.name,
                     metadata={
                         'title': 'name',
                         "type": "string",
                         'description': 'The inverter name',
                         'readOnly': True,
                     }))

        self.serial = Value(inverter.serial)
        self.add_property(
            Property(self,
                     'serial',
                     self.serial,
                     metadata={
                         'title': 'serial',
                         "type": "string",
                         'description': 'The serial number',
                         'readOnly': True,
                     }))

        self.available = Value(inverter.is_available)
        self.add_property(
            Property(self,
                     'is_available',
                     self.available,
                     metadata={
                         'title': 'available',
                         "type": "boolean",
                         'description': 'True, if is available',
                         'readOnly': True,
                     }))

        self.producing = Value(inverter.is_producing)
        self.add_property(
            Property(self,
                 'is_producing',
                     self.producing,
                     metadata={
                     'title': 'producing',
                     "type": "boolean",
                     'description': 'True, if is producing',
                     'readOnly': True,
                 }))

        self.power_max = Value(inverter.power_max)
        self.add_property(
            Property(self,
                     'power_max',
                     self.power_max,
                     metadata={
                         'title': 'power_max',
                         "type": "integer",
                         'description': 'The max power [W]',
                         'readOnly': True,
                     }))

        self.power_limit = Value(inverter.power_limit, inverter.set_power_limit)
        self.add_property(
            Property(self,
                     'power_limit',
                     self.power_limit,
                     metadata={
                         'title': 'power_limit',
                         "type": "integer",
                         'description': 'The power limit [W]',
                         'readOnly': False,
                     }))

        self.spare_power = Value(inverter.spare_power)
        self.add_property(
            Property(self,
                     'spare_power',
                     self.spare_power,
                     metadata={
                         'title': 'spare_power',
                         "type": "integer",
                         'description': 'The potential spare power [W]',
                         'readOnly': True,
                     }))

        self.temp = Value(inverter.temp)
        self.add_property(
            Property(self,
                     'temp',
                     self.temp,
                     metadata={
                         'title': 'temperature',
                         "type": "string",
                         'description': 'The temperature [C]',
                         'readOnly': True,
                     }))

        self.u_dc1 = Value(inverter.u_dc1)
        self.add_property(
            Property(self,
                     'u_dc1',
                     self.u_dc1,
                     metadata={
                         'title': 'Voltage DC - Channel 1',
                         "type": "number",
                         'description': 'The voltage DC channel 1 [V]',
                         'readOnly': True,
                     }))

        self.u_dc2 = Value(inverter.u_dc2)
        self.add_property(
            Property(self,
                     'u_dc2',
                     self.u_dc2,
                     metadata={
                         'title': 'Voltage DC - Channel 2',
                         "type": "number",
                         'description': 'The voltage DC channel 2 [V]',
                         'readOnly': True,
                     }))

        self.i_dc1 = Value(inverter.i_dc1)
        self.add_property(
            Property(self,
                     'i_dc1',
                     self.i_dc1,
                     metadata={
                         'title': 'Current DC - Channel 1',
                         "type": "number",
                         'description': 'The current DC channel 1 [A]',
                         'readOnly': True,
                     }))

        self.i_dc2 = Value(inverter.i_dc2)
        self.add_property(
            Property(self,
                     'i_dc2',
                     self.i_dc2,
                     metadata={
                         'title': 'Current DC - Channel 2',
                         "type": "number",
                         'description': 'The current DC channel 2 [A]',
                         'readOnly': True,
                     }))

        self.p_dc = Value(inverter.p_dc)
        self.add_property(
            Property(self,
                     'p_dc',
                     self.p_dc,
                     metadata={
                         'title': 'Power DC',
                         "type": "number",
                         'description': 'The power DC [W]',
                         'readOnly': True,
                     }))

        self.p_dc1 = Value(inverter.p_dc1)
        self.add_property(
            Property(self,
                     'p_dc1',
                     self.p_dc1,
                     metadata={
                         'title': 'Power DC channel 1',
                         "type": "number",
                         'description': 'The power DC channel 1 [W]',
                         'readOnly': True,
                     }))

        self.p_dc2 = Value(inverter.p_dc2)
        self.add_property(
            Property(self,
                     'p_dc2',
                     self.p_dc2,
                     metadata={
                         'title': 'Power DC channel 2',
                         "type": "number",
                         'description': 'The power DC channel 2 [W]',
                         'readOnly': True,
                     }))

        self.p_ac = Value(inverter.p_ac)
        self.add_property(
            Property(self,
                     'p_ac',
                     self.p_ac,
                     metadata={
                         'title': 'Power AC',
                         "type": "number",
                         'description': 'The power AC [W]',
                         'readOnly': True,
                     }))

        self.u_ac = Value(inverter.u_ac)
        self.add_property(
            Property(self,
                     'u_ac',
                     self.u_ac,
                     metadata={
                         'title': 'Voltage AC',
                         "type": "number",
                         'description': 'The voltage AC [V]',
                         'readOnly': True,
                     }))

        self.i_ac = Value(inverter.i_ac)
        self.add_property(
            Property(self,
                     'i_ac',
                     self.i_ac,
                     metadata={
                         'title': 'Current AC',
                         "type": "number",
                         'description': 'The current AC [A]',
                         'readOnly': True,
                     }))

        self.efficiency = Value(inverter.efficiency)
        self.add_property(
            Property(self,
                     'efficiency',
                     self.efficiency,
                     metadata={
                         'title': 'Efficiency',
                         "type": "number",
                         'description': 'The efficiency [%]',
                         'readOnly': True,
                     }))

        self.frequency = Value(inverter.frequency)
        self.add_property(
            Property(self,
                     'frequency',
                     self.frequency,
                     metadata={
                         'title': 'frequency',
                         "type": "number",
                         'description': 'The frequency [Hz]',
                         'readOnly': True,
                     }))


        self.ioloop = tornado.ioloop.IOLoop.current()
        self.inverter.register_listener(self.on_value_changed)

    def on_value_changed(self, inverter: Inverter):
        self.ioloop.add_callback(self.__on_value_changed)

    def __on_value_changed(self):
        self.producing.notify_of_external_update(self.inverter.is_producing)
        self.available.notify_of_external_update(self.inverter.is_available)
        self.p_dc.notify_of_external_update(self.inverter.p_dc)
        self.p_dc1.notify_of_external_update(self.inverter.p_dc1)
        self.p_dc2.notify_of_external_update(self.inverter.p_dc2)
        self.u_dc1.notify_of_external_update(self.inverter.u_dc1)
        self.u_dc2.notify_of_external_update(self.inverter.u_dc2)
        self.i_dc1.notify_of_external_update(self.inverter.i_dc1)
        self.i_dc2.notify_of_external_update(self.inverter.i_dc2)
        self.p_ac.notify_of_external_update(self.inverter.p_ac)
        self.i_ac.notify_of_external_update(self.inverter.i_ac)
        self.u_ac.notify_of_external_update(self.inverter.u_ac)
        self.frequency.notify_of_external_update(self.inverter.frequency)
        self.efficiency.notify_of_external_update(self.inverter.efficiency)
        self.timestamp_last_success.notify_of_external_update(self.inverter.timestamp_last_success.strftime("%Y-%m-%dT%H:%M:%S"))
        self.temp.notify_of_external_update(self.inverter.temp)
        self.power_max.notify_of_external_update(self.inverter.power_max)
        self.power_limit.notify_of_external_update(self.inverter.power_limit)
        self.spare_power.notify_of_external_update(self.inverter.spare_power)


def run_server(description: str, port: int, base_uri: str):
    awning_webthings = [InverterWebThing(description, inverter) for inverter in  Dtu.connect(base_uri).inverters]
    server = WebThingServer(MultipleThings(awning_webthings, 'Inverters'), port=port, disable_host_validation=True)

    logging.info('running webthing server http://localhost:' + str(port))
    try:
        server.start()
    except KeyboardInterrupt:
        logging.info('stopping webthing server')
        server.stop()
        logging.info('done')




from threading import Thread
from random import randint
from typing import Set, Optional
import re
import requests
from requests import Session
import logging
from time import sleep
from datetime import datetime


class Inverter:

    @staticmethod
    def connect(base_uri: str, inverter_name: str) -> Optional:
        dtu = Dtu.connect(base_uri, {inverter_name})
        return None if len(dtu.inverters) == 0 else dtu.inverters[0]

    def __init__(self, base_uri: str, id: int, channels: int, name: str, serial: str, interval: int):
        self.is_running = True
        self.uri = base_uri
        self.update_uri = re.sub("^/|/$", "", base_uri) + '/api/ctrl'
        self.live_uri = re.sub("^/|/$", "", base_uri) + '/api/record/live'
        self.index_uri = re.sub("^/|/$", "", base_uri) + '/api/index'
        self.config_uri = re.sub("^/|/$", "", base_uri) + '/api/record/config'
        self.inverter_uri = re.sub("^/|/$", "", base_uri) + '/api/inverter/list'

        self.id = id
        self.channel = channels
        self.name = name
        self.serial = serial
        self.interval = interval
        self.p_dc = 0
        self.p_dc1 = 0
        self.p_dc2 = 0
        self.u_dc1 = 0
        self.u_dc2 = 0
        self.i_dc1 = 0
        self.i_dc2 = 0
        self.p_ac = 0
        self.u_ac = 0
        self.i_ac = 0
        self.temp = 0
        self.frequency = 0
        self.efficiency = 0
        self.power_max = 0
        self.power_limit = 0
        self.timestamp_last_success = datetime.fromtimestamp(0)
        self.is_available = False
        self.is_producing = False
        self.listener = None
        self.session = Session()
        Thread(target=self.__periodic_refresh, daemon=True).start()

    @property
    def spare_power(self) -> int:
        if self.p_ac < (self.power_limit * 0.9):
            return 0
        else:
            return self.power_max - self.power_limit

    def close(self):
        self.is_running = False

    def __renew_session(self):
        try:
            self.session.close()
        except Exception as e:
            logging.warning("error occurred closing session " + str(e))
        self.session = Session()

    def __periodic_refresh(self):
        while self.is_running:
            try:
                sleep(randint(0, self.interval))
                self.refresh()
                sleep(int(self.interval/5))
            except Exception as e:
                logging.warning("error occurred refreshing inverter " + self.name + " " + str(e) + " (max " + str(self.power_max) + " watt)")
                sleep(5)
                try:
                    self.__renew_session()
                except Exception as e:
                    logging.warning("error occurred renewing session " + str(e))

    def refresh(self):
        try:
            # fetch inverter info
            response = self.session.get(self.index_uri, timeout=60)
            response.raise_for_status()
            inverter_state = response.json()['inverter']

            timestamp_last_success = datetime.fromtimestamp(inverter_state[self.id]['ts_last_success'])

            previous_is_available = self.is_available
            self.is_available = inverter_state[self.id]['is_avail']
            if previous_is_available != self.is_available:
                logging.info("inverter " + str(self.name) + " is " + ("" if self.is_available else "not ") + "available")

            previous_is_producing = self.is_producing
            self.is_producing = inverter_state[self.id]['is_producing']
            if previous_is_producing != self.is_producing:
                logging.info("inverter " + str(self.name) + " is " + ("" if self.is_producing else "not ") + "producing")

            if self.is_producing:
                # fetch power limit
                response = self.session.get(self.config_uri, timeout=60)
                response.raise_for_status()

                inverter_configs = response.json()['inverter']

                # fetch inverter info
                response = self.session.get(self.inverter_uri, timeout=60)
                response.raise_for_status()
                inverter_infos = response.json()['inverter']

                # fetch temp, power, etc
                response = self.session.get(self.live_uri, timeout=60)
                response.raise_for_status()
                inverter_measures = response.json()['inverter']

                p_ac = 0
                i_ac = 0
                u_ac  =0
                p_dc = 0
                p_dc1 = None
                p_dc2 = None
                u_dc1 = None
                u_dc2 = None
                i_dc1 = None
                i_dc2 = None
                efficiency = None
                temp = 0
                frequency = 0
                power_limit = 0
                power_max = sum(inverter_infos[self.id]['ch_max_pwr'])

                for config in inverter_configs[self.id]:
                    if config['fld'] == 'active_PowerLimit':
                        power_limit_percent = float(config['val'])
                        power_limit = int(power_max * power_limit_percent / 100)

                for measure in inverter_measures[self.id]:
                    if measure['fld'] == 'P_AC':
                        p_ac = float(measure['val'])
                    elif measure['fld'] == 'I_AC':
                        i_ac = float(measure['val'])
                    elif measure['fld'] == 'U_AC':
                        u_ac = float(measure['val'])
                    elif measure['fld'] == 'U_DC':
                        if u_dc1 is None:
                            u_dc1 = float(measure['val'])
                        else:
                            u_dc2 = float(measure['val'])
                    elif measure['fld'] == 'I_DC':
                        if i_dc1 is None:
                            i_dc1 = float(measure['val'])
                        else:
                            i_dc2 = float(measure['val'])
                    elif measure['fld'] == 'P_DC':
                        if p_dc1 is None:
                            p_dc1 = float(measure['val'])
                        elif p_dc2 is None:
                            p_dc2 = float(measure['val'])
                        else:
                            p_dc = float(measure['val'])
                    elif measure['fld'] == 'Efficiency':
                        efficiency = float(measure['val'])
                    elif measure['fld'] == 'Temp':
                        temp = float(measure['val'])
                    elif measure['fld'] == 'F_AC':
                        frequency = float(measure['val'])

                self.update(timestamp_last_success, power_max, power_limit, p_ac, u_ac, i_ac, p_dc, p_dc1, p_dc2, u_dc1, u_dc2, i_dc1, i_dc2, efficiency, temp, frequency)
            else:
                self.update(timestamp_last_success, self.power_max, self.power_limit, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)

        except Exception as e:
            logging.warning("error occurred getting " + self.name + " inverter data " + str(e))

    def set_power_limit(self, limit_watt: int):
        logging.info("inverter " + self.name + " setting (non-persistent) absolute power limit to " + str(limit_watt) + " Watt")
        try:
            data = {"id": self.id,
                    "cmd": "limit_nonpersistent_absolute",
                    "val": limit_watt}
            resp = requests.post(self.update_uri, json=data)
            resp.raise_for_status()
        except Exception as e:
            logging.warning("error occurred updating power limit of " + self.name + " inverter with " + str(limit_watt) + " " + str(e))

    def update(self, timestamp_last_success: datetime, power_max: int, power_limit: int, p_ac: float, u_ac: float, i_ac: float, p_dc: float, p_dc1:float, p_dc2: float, u_dc1: float, u_dc2: float, i_dc1: float, i_dc2: float, efficiency: float, temp: float, frequency: float):
        if timestamp_last_success != self.timestamp_last_success:
            self.timestamp_last_success = timestamp_last_success
            self.power_max = power_max
            self.power_limit = power_limit
            if self.p_ac != p_ac:
                self.p_ac = p_ac
                logging.debug("inverter " + self.name + " current power: " + str(self.p_ac) + " Watt " +
                              "(power limit " + str(self.power_limit) + " Watt; spare power " + str(self.spare_power) + " Watt)")
            self.u_ac = u_ac
            self.u_dc1 = u_dc1
            self.u_dc2 = u_dc2
            self.i_dc1 = i_dc1
            self.i_dc2 = i_dc2
            self.i_ac = i_ac
            self.p_dc = p_dc
            self.p_dc1 = p_dc1
            self.p_dc2 = p_dc2
            self.efficiency = efficiency
            self.temp = temp
            self.frequency = frequency
            self.__notify_listener()

    def register_listener(self, listener):
        self.listener = listener

    def __notify_listener(self):
        if self.listener is not None:
            self.listener(self)

    def __str__(self):
        return self.name + " " + self.serial + " (P_AC: " + str(self.p_ac) + ", U_AC: " + str(self.u_ac) + ", I_AC: " + str(self.i_ac) + \
                ", P_DC: " + str(self.p_dc) + ", EFFICIENCY: " + str(self.efficiency) +  ")"

    def __repr__(self):
        return  self.__str__()



class Dtu:

    def __init__(self, base_uri: str, inverter_filter: Set[str]):
        self.base_uri = base_uri
        uri = re.sub("^/|/$", "", self.base_uri) + '/api/inverter/list'
        response = requests.get(uri)
        data = response.json()
        interval = int(data['interval'])
        self.inverters = [Inverter(self.base_uri, entry['id'], entry['channels'], entry['name'], entry['serial'], interval)
                          for entry in data['inverter']
                          if len(inverter_filter) == 0 or entry['name'] in inverter_filter]

    def inverter_by_name(self, name: str) -> Optional[Inverter]:
        for inverter in self.inverters:
            if inverter.name == name:
                return inverter
        return None

    @staticmethod
    def connect(base_uri: str, inverter_filter: Set[str] = None):
        return Dtu(base_uri, set() if inverter_filter is None else inverter_filter)

    def close(self):
        for inverter in self.inverters:
            inverter.close()

